/*
 * Copyright (c) 2018-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct  4 2018
 */
#include "acgps.h"

#include <sys/time.h>

#include <err.h>
#include <math.h>
#include <poll.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "gps_c_types.h"
#include "codels.h"


/* --- Available hardware drivers ------------------------------------------ */

extern const struct gps_driver_s novatel_driver;
extern const struct gps_driver_s tersus_driver;
extern const struct gps_driver_s ublox_driver;

static const struct gps_driver_s *gps_drivers[] = {
  &novatel_driver,
  &tersus_driver,
  &ublox_driver,
  NULL
};


/* --- Task io ---------------------------------------------------------- */


/** Codel gpsio_start of task io.
 *
 * Triggered by gps_start.
 * Yields to gps_poll.
 * Throws gps_e_sys.
 */
genom_event
gpsio_start(gps_ids *ids, const genom_context self)
{
  /* initialize default ENU reference (LAAS) */
  gps_set_reference(&(gps_llh_s){
      .latitude = 43.56221677048,
        .longitude = 1.47730192335,
        .height = 206.232826165
        }, &ids->ref, self);
  ids->ref_sigma = (gps_enu_s){
    .north = 0.1165,
    .east = 0.0926,
    .up = 0.2156
  };

  ids->fix = (gps_ids_fix_s){
    .remain = 0., .samples = 0,
    .llh = { .latitude = nan(""), .longitude = nan(""), .height = nan("") },
    .pos = { .east = nan(""), .north = nan(""), .up = nan("") },
    .sigma = { .east = nan(""), .north = nan(""), .up = nan("") }
  };

  /* initialize a disconnected device */
  ids->device = malloc(sizeof(*ids->device));
  if (!ids->device) return gps_e_sys_error(NULL, self);

  ids->device->channel.fd = -1;
  ids->device->channel.r = ids->device->channel.w = 0;

  ids->device->driver = NULL;
  ids->device->data = NULL;

  ids->solution = (gps_solution_s){
    .fix = gps_fix_none
  };

  ids->devicefd = (gps_devicefd_s *)&ids->device->channel.fd;

  ids->param = (gps_param_s){
    .model = "",
    .rate = 0.,
    .ecutoff = nan("")
  };

  /* rtk connection */
  ids->rtk_port = 8083;
  ids->rtk_conn = malloc(sizeof(*ids->rtk_conn));
  if (!ids->rtk_conn) return gps_e_sys_error(NULL, self);
  *ids->rtk_conn = (gps_rtk_conn_s){
    .fd4 = -1, .fd6 = -1,
    .clients = NULL, .servers = NULL
  };

  if (gpsrtk_conn_open(ids->rtk_conn, ids->rtk_port))
    warn(NULL);

  /* logging */
  ids->log = malloc(sizeof(*ids->log));
  if (!ids->log) abort();
  *ids->log = (gps_log_s){
    .req = {
      .aio_fildes = -1,
      .aio_offset = 0,
      .aio_buf = ids->log->buffer,
      .aio_nbytes = 0,
      .aio_reqprio = 0,
      .aio_sigevent = { .sigev_notify = SIGEV_NONE },
      .aio_lio_opcode = LIO_NOP
    },
    .pending = false, .skipped = false,
    .decimation = 1, .missed = 0, .total = 0
  };

  return gps_poll;
}


/** Codel gpsio_poll of task io.
 *
 * Triggered by gps_poll.
 * Yields to gps_poll, gps_recv, gps_rtk, gps_err.
 * Throws gps_e_sys.
 */
genom_event
gpsio_poll(const gps_devicefd_s *devicefd, gps_rtk_conn_s **rtk_conn,
           const genom_context self)
{
  (void)self;

  struct pollfd pfd[] = {
    { .fd = devicefd->fd, .events = POLLIN },
    { .fd = (*rtk_conn)->fd4, .events = POLLIN },
    { .fd = (*rtk_conn)->fd6, .events = POLLIN }
  };
  int s;

  gpsrtk_keepalive(*rtk_conn);

  s = poll(pfd, sizeof(pfd)/sizeof(pfd[0]), 500/*ms*/);
  if (s < 0) return gps_err;
  if (s == 0) return gps_poll;

  if (pfd[0].revents & POLLHUP) return gps_err;
  if (pfd[0].revents & POLLIN) return gps_recv;
  if (pfd[1].revents & POLLIN) return gps_rtk;
  if (pfd[2].revents & POLLIN) return gps_rtk;

  return gps_err;
}


/** Codel gpsio_dev_recv of task io.
 *
 * Triggered by gps_recv.
 * Yields to gps_poll, gps_publish.
 * Throws gps_e_sys.
 */
genom_event
gpsio_dev_recv(gps_device_s **device, gps_solution_s *solution,
               sequence_octet *rtk_out, const genom_context self)
{
  (void)self;
  int s;

  if ((*device)->driver && (*device)->driver->decode) {
    s = (*device)->driver->decode(*device, solution, rtk_out);
    switch(s) {
      case 0: return gps_publish;
      case EAGAIN: break;
      default:
        warn("receiving data");
    }
  }

  return gps_poll;
}


/** Codel gpsio_rtk_recv of task io.
 *
 * Triggered by gps_rtk.
 * Yields to gps_poll.
 * Throws gps_e_sys.
 */
genom_event
gpsio_rtk_recv(gps_device_s **device, gps_rtk_conn_s **rtk_conn,
               sequence_octet *rtk_in,
               const sequence_unsigned_short *rtk_ignore,
               const genom_context self)
{
  (void)self;
  const uint8_t *match;
  uint16_t type;
  size_t len;
  size_t i;

  if (gpsrtk_recv(*rtk_conn, rtk_in)) return gps_poll;
  if (!rtk_in->_length) return gps_poll;

  while(gpsrtk_match_rtcm(rtk_in->_buffer, rtk_in->_length,
                          &match, &type, &len)) {
    /* ignore unwanted messages */
    for (i = 0; i < rtk_ignore->_length; i++)
      if (rtk_ignore->_buffer[i] == type) { type = 0; break; }

    if (type && (*device)->driver) {
      if ((*device)->driver->ctl(*device, GPSCTL_INRTCM, type, match, len))
        warnx("ignoring RTCM %d data", type);
    }

    rtk_in->_length -= match - rtk_in->_buffer + len;
    memmove(rtk_in->_buffer, match + len, rtk_in->_length);
  }

  /* trim any unrecognized dandling old data */
  static const size_t tlen = 10240;
  if (rtk_in->_length > 2 * tlen) {
    warnx("skipping bad %zu RTCM bytes", tlen);
    memmove(rtk_in->_buffer, rtk_in->_buffer + tlen, rtk_in->_length - tlen);
  }

  return gps_poll;
}


/** Codel gpsio_publish of task io.
 *
 * Triggered by gps_publish.
 * Yields to gps_poll.
 * Throws gps_e_sys.
 */
genom_event
gpsio_publish(const gps_solution_s *solution,
              gps_rtk_conn_s **rtk_conn, sequence_octet *rtk_out,
              const gps_info *info, const gps_state *state,
              gps_log_s **log, const genom_context self)
{
  (void)state;
  gps_solution_s *info_data = info->data(self);
  or_pose_estimator_state *state_data = state->data(self);
  gps_log_s *log_data = *log;

  /* publish rtcm */
  gpsrtk_send(*rtk_conn, rtk_out);

  /* publish solution & state */
  if (info_data->utc.nsec == solution->utc.nsec &&
      info_data->utc.sec == solution->utc.sec)
    return gps_poll; /* no update */

  *info_data = *solution;
  info->write(self);

  *state_data = (or_pose_estimator_state){
    .ts = solution->ts,
    .intrinsic = false,

    .pos = {
      ._present = solution->fix == gps_fix_none ? false : true,
      ._value = {
        .x = solution->pos.north,
        .y = - solution->pos.east,
        .z = solution->pos.up
      }
    },
    .pos_cov = {
      ._present = solution->fix == gps_fix_none ? false : true,
      ._value.cov = {
        [0] = solution->pos_sigma.north * solution->pos_sigma.north,
        [2] = solution->pos_sigma.east * solution->pos_sigma.east,
        [5] = solution->pos_sigma.up * solution->pos_sigma.up
      }
    },

    .vel = {
      ._present = !isnan(solution->vel.north),
      ._value = {
        .vx = solution->vel.north,
        .vy = - solution->vel.east,
        .vz = solution->vel.up
      }
    },
    .vel_cov = {
      ._present = !isnan(solution->vel.north),
      ._value.cov = {
        [0] = solution->vel_sigma.north * solution->vel_sigma.north,
        [2] = solution->vel_sigma.east * solution->vel_sigma.east,
        [5] = solution->vel_sigma.up * solution->vel_sigma.up
      }
    }
  };
  state->write(self);

  /* logging */
  if (log_data->req.aio_fildes >= 0) {
    log_data->total++;
    if (log_data->total % log_data->decimation == 0) {
      if (log_data->pending) {
        if (aio_error(&log_data->req) != EINPROGRESS) {
          log_data->pending = false;
          if (aio_return(&log_data->req) <= 0) {
            warn("log");
            close(log_data->req.aio_fildes);
            log_data->req.aio_fildes = -1;
          }
        } else {
          log_data->skipped = true;
          log_data->missed++;
        }
      }
    }

    if (log_data->req.aio_fildes >= 0 && !log_data->pending) {
      log_data->req.aio_nbytes = snprintf(
        log_data->buffer, sizeof(log_data->buffer),
        "%s" gps_log_fmt "\n",
        log_data->skipped ? "\n" : "",
        solution->ts.sec, solution->ts.nsec,
        solution->utc.sec, solution->utc.nsec,
        gps_fixstr(solution->fix), solution->fix, solution->sats,
        solution->llh.latitude, solution->llh.longitude, solution->llh.height,
        solution->pos.east, solution->pos.north, solution->pos.up,
        solution->vel.east, solution->vel.north, solution->vel.up,
        solution->pos_sigma.east,
        solution->pos_sigma.north,
        solution->pos_sigma.up,
        solution->vel_sigma.east,
        solution->vel_sigma.north,
        solution->vel_sigma.up);

      if (aio_write(&log_data->req)) {
        warn("log");
        close(log_data->req.aio_fildes);
        log_data->req.aio_fildes = -1;
      } else
        log_data->pending = true;

      log_data->skipped = false;
    }
  }

  return gps_poll;
}


/** Codel gpsio_close of task io.
 *
 * Triggered by gps_err.
 * Yields to gps_poll.
 * Throws gps_e_sys.
 */
genom_event
gpsio_close(gps_device_s **device, gps_param_s *param,
            const genom_context self)
{
  (void)self;

  (*device)->driver = NULL;
  gpstty_close(&(*device)->channel);
  if ((*device)->data) free((*device)->data);
  (*device)->data = NULL;

  *param = (gps_param_s){
    .model = "",
    .rate = 0.,
    .ecutoff = nan("")
  };

  return gps_poll;
}


/** Codel gpsio_stop of task io.
 *
 * Triggered by gps_stop.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gpsio_stop(gps_device_s **device, gps_param_s *param,
           gps_rtk_conn_s **rtk_conn, const genom_context self)
{
  gps_disconnect_start(device, param, self);
  free(*device);

  gpsrtk_conn_close(*rtk_conn);
  free(*rtk_conn);

  return gps_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel gps_connect_start of activity connect.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev, gps_e_baddev.
 */
genom_event
gps_connect_start(const char serial[64], uint32_t baud,
                  gps_device_s **device, gps_param_s *param,
                  const genom_context self)
{
  uint32_t baud_list[] = {
    baud,
    9600, 38400, 57600, 115200, 230400, 460800, 921600,
    0
  };
  int i, d;

  /* open tty */
  gpsio_close(device, param, self);
  if (gpstty_open(&(*device)->channel, serial))
    return gps_e_sys_error(serial, self);

  /* check endpoint */
  for(i = 0; !(*device)->data && baud_list[i]; i++) {
    if (i > 0 && baud_list[i] == baud_list[0]) continue;

    if (gpstty_setspeed(&(*device)->channel, baud_list[i]))
      return gps_e_sys_error(serial, self);

    for(d = 0; gps_drivers[d]; d++) {
      (*device)->data = gps_drivers[d]->match(&(*device)->channel);
      if (!(*device)->data) continue;
      (*device)->driver = gps_drivers[d];

      warnx("detected %s protocol at %dbps",
            gps_drivers[d]->protocol, baud_list[i]);
      if (i == 0) { /* matched at required baud rate */ break; }

      /* configure required baud rate */
      gps_drivers[d]->ctl(*device, GPSCTL_BAUD, baud);
      poll(NULL, 0, 1000); /* let the device rest a bit (heuristic delay) */

      if (gpstty_setspeed(&(*device)->channel, baud))
        return gps_e_sys_error(serial, self);

      /* match again or fail */
      free((*device)->data);
      (*device)->data = gps_drivers[d]->match(&(*device)->channel);
      if (!(*device)->data) {
        warnx("failed to reconfigure at %dbps", baud);
        gpsio_close(device, param, self);
        return gps_e_nodev(self);
      }
      break;
    }
  }
  if (!(*device)->data) {
    warnx("no matched device");
    gpsio_close(device, param, self);
    return gps_e_nodev(self);
  }

  if ((*device)->driver->configure(*device)) {
    warnx("failed to configure device");

    gpsio_close(device, param, self);
    return gps_e_nodev(self);
  }

  if ((*device)->driver->ctl(*device, GPSCTL_Q_PARAM, param)) {
    warnx("failed to query device settings");

    gpsio_close(device, param, self);
    return gps_e_nodev(self);
  }

  warnx("connected to %s at %dbps (%gHz updates) on %s",
        param->model, (*device)->channel.baud,
        param->rate, (*device)->channel.path);

  return gps_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel gps_disconnect_start of activity disconnect.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gps_disconnect_start(gps_device_s **device, gps_param_s *param,
                     const genom_context self)
{
  if ((*device)->driver)
    if ((*device)->driver->unconfigure(*device))
      warnx("failed to unconfigure device");

  gpsio_close(device, param, self);
  return gps_ether;
}


/* --- Activity set_rtk_port -------------------------------------------- */

/** Codel gps_set_rtk_port of activity set_rtk_port.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gps_set_rtk_port(gps_rtk_conn_s **rtk_conn, uint16_t port,
                 uint16_t *rtk_port, const genom_context self)
{
  if (gpsrtk_conn_open(*rtk_conn, port))
    return gps_e_sys_error(NULL, self);

  *rtk_port = port;
  return gps_ether;
}


/* --- Activity connect_rtk --------------------------------------------- */

/** Codel gps_connect_rtk_start of activity connect_rtk.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev.
 */
genom_event
gps_connect_rtk_start(const char host[64], uint16_t port,
                      gps_rtk_conn_s **rtk_conn,
                      const genom_context self)
{
  int s;

  s = gpsrtk_server_open(host, port, *rtk_conn);
  if (s) return gps_e_nodev(self);

  return gps_ether;
}


/* --- Activity unfix --------------------------------------------------- */

/** Codel gps_unfix_start of activity unfix.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev.
 */
genom_event
gps_unfix_start(gps_device_s **device, const genom_context self)
{
  int s;

  if (!(*device)->driver) return gps_e_nodev(self);

  s = (*device)->driver->ctl(*device, GPSCTL_UNFIX);
  if (s) { errno = EIO; return gps_e_sys_error("unfix", self); }

  return gps_ether;
}


/* --- Activity fix ----------------------------------------------------- */

/** Codel gps_fix_set of activity fix.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev, gps_e_nofix.
 */
genom_event
gps_fix_set(gps_device_s **device, const gps_llh_s *llh,
            const gps_enu_s *sigma, gps_llh_s *ref,
            gps_enu_s *ref_sigma, const genom_context self)
{
  int s;

  /* require a bit of accuracy */
  if (sigma->east > 2. || sigma->north > 2. || sigma->up > 2.)
    return gps_e_nofix(self);

  s = (*device)->driver->ctl(*device, GPSCTL_FIX, llh);
  if (s) { errno = EIO; return gps_e_sys_error("fix", self); }

  /* set ENU reference here */
  *ref_sigma = *sigma;
  gps_set_reference(llh, ref, self);

  return gps_ether;
}


/* --- Activity fix_here ------------------------------------------------ */

/** Codel gps_fix_here_start of activity fix_here.
 *
 * Triggered by gps_start.
 * Yields to gps_avg, gps_ether.
 * Throws gps_e_sys, gps_e_nodev, gps_e_nofix.
 */
genom_event
gps_fix_here_start(gps_device_s **device, double avg_time,
                   uint16_t *delay, gps_ids_fix_s *fix,
                   or_time_ts *last, or_time_ts *deadline,
                   const genom_context self)
{
  struct timeval tv;
  genom_event e;

  /* compute absolute deadline */
  if (avg_time < 0.) {
    errno = EINVAL;
    return gps_e_sys_error("avg_time", self);
  }
  gettimeofday(&tv, NULL);
  tv.tv_sec += floor(avg_time);
  tv.tv_usec += (avg_time - floor(avg_time)) * 1000000;
  if (tv.tv_usec > 1000000) {
    tv.tv_sec += tv.tv_usec / 1000000;
    tv.tv_usec = tv.tv_usec % 1000000;
  }

  deadline->sec = tv.tv_sec;
  deadline->nsec = tv.tv_usec * 1000;

  /* delay between samples:
   * 1s, 1min or 10min depending on the total average time */
  if (avg_time <= 60) *delay = 1;
  else if (avg_time <= 3600) *delay = 60;
  else *delay = 600;

  /* unset fixed position */
  e = gps_unfix_start(device, self);
  if (e != gps_ether) return e;

  *fix = (gps_ids_fix_s){
    .remain = avg_time, .samples = 0,
    .llh = { 0. },
    .pos = { 0. },
    .sigma = { .east = 1e7, .north = 1e7, .up = 1e7 }
  };
  *last = (or_time_ts){ 0 };

  return gps_avg;
}

/** Codel gps_fix_here_avg of activity fix_here.
 *
 * Triggered by gps_avg.
 * Yields to gps_pause_avg, gps_stop.
 * Throws gps_e_sys, gps_e_nodev, gps_e_nofix.
 */
genom_event
gps_fix_here_avg(uint16_t delay, const or_time_ts *deadline,
                 const gps_solution_s *solution, or_time_ts *last,
                 gps_ids_fix_s *fix, const genom_context self)
{
  (void)self;
  struct timeval tv;

  /* average only new valid positions after sampling delay */
  if (solution->utc.sec >= last->sec + delay &&
      solution->fix > gps_fix_interp && solution->fix < gps_fix_fixed) {
    *last = solution->utc;

    gpsconv_llhavg(
      &solution->llh, &solution->pos_sigma, &fix->llh, &fix->sigma);
    gpsconv_llhenu(&fix->llh, &fix->pos);
    fix->samples++;
  }

  gettimeofday(&tv, NULL);
  fix->remain = deadline->sec - tv.tv_sec;
  fix->remain += 1e-9 * (deadline->nsec - 1000 * tv.tv_usec);
  if (fix->remain >= 0.) return gps_pause_avg;

  fix->remain = 0.;
  return gps_stop;
}

/** Codel gps_fix_done of activity fix_here.
 *
 * Triggered by gps_stop.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev, gps_e_nofix.
 */
genom_event
gps_fix_done(gps_device_s **device, gps_ids_fix_s *fix, gps_llh_s *llh,
             gps_enu_s *pos, gps_enu_s *sigma, gps_llh_s *ref,
             gps_enu_s *ref_sigma, const genom_context self)
{
  *llh = fix->llh;
  *pos = fix->pos;
  *sigma = fix->sigma;

  return gps_fix_set(device, &fix->llh, &fix->sigma, ref, ref_sigma, self);
}


/* --- Activity send_rtcm ----------------------------------------------- */

/** Codel gps_send_rtcm of activity send_rtcm.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev.
 */
genom_event
gps_send_rtcm(gps_device_s **device, uint16_t rtcmid, double period,
              double offset, sequence_unsigned_short *rtk_ignore,
              const genom_context self)
{
  size_t i;
  int s;

  if (!(*device)->driver) return gps_e_nodev(self);
  if (period <= 0.) {
    errno = EINVAL; return gps_e_sys_error("send_rtcm", self);
  }

  /* configure hardware to stream this message */
  s = (*device)->driver->ctl(*device, GPSCTL_OUTRTCM, rtcmid, period, offset);
  if (s) { errno = EIO; return gps_e_sys_error("send_rtcm", self); }

  /* remove from the list of ignored messages */
  for (i = 0; i < rtk_ignore->_length; i++)
    if (rtk_ignore->_buffer[i] == rtcmid) {
      memmove(
        &rtk_ignore->_buffer[i], &rtk_ignore->_buffer[i+1],
        (rtk_ignore->_length - i - 1) * sizeof(*rtk_ignore->_buffer));
      rtk_ignore->_length--;
      break;
    }

  return gps_ether;
}


/* --- Activity cancel_rtcm --------------------------------------------- */

/** Codel gps_cancel_rtcm of activity cancel_rtcm.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys, gps_e_nodev.
 */
genom_event
gps_cancel_rtcm(gps_device_s **device, uint16_t rtcmid,
                sequence_unsigned_short *rtk_ignore,
                const genom_context self)
{
  size_t i;
  int s;

  if (!(*device)->driver) return gps_e_nodev(self);

  /* append to the list of ignored incoming messages */
  for (i = 0; i < rtk_ignore->_length; i++)
    if (rtk_ignore->_buffer[i] == rtcmid) break;

  if (i >= rtk_ignore->_length) {
    if (genom_sequence_reserve(rtk_ignore, i + 1))
      return gps_e_sys_error("cancel_rtcm", self);
    rtk_ignore->_buffer[i] = rtcmid;
    rtk_ignore->_length = i + 1;
  }

  /* configure hardware to not stream this message */
  s = (*device)->driver->ctl(*device, GPSCTL_OUTRTCM, rtcmid, 0., 0.);
  if (s) { errno = EIO; return gps_e_sys_error("cancel_rtcm", self); }

  return gps_ether;
}


/* --- Activity monitor_fix --------------------------------------------- */

/** Codel gps_monitor_fix_start of activity monitor_fix.
 *
 * Triggered by gps_start.
 * Yields to gps_pause_start, gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gps_monitor_fix_start(gps_monitor_op_s trigger, gps_fix_s target,
                      gps_fix_s fix, const genom_context self)
{
  if (target > gps_fix_none)
    switch (trigger) {
      case gps_monitor_above:
        return (fix < target) ? gps_pause_start : gps_ether;

      case gps_monitor_below:
        return (fix >= target) ? gps_pause_start : gps_ether;
    }

  errno = EINVAL;
  return gps_e_sys_error("monitor_fix", self);
}
